/**
 * The History class represents a kobold's history.
 * In essence, it's a key-value store to be used by scenes to store
 * values at the time of reading, and to log choices made.
 */
function History() {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    // PUBLIC MEMBERS \\        this.public = "test";

    // Scene-specific storage.
    this.scenes = {};
    // Shared storage.
    this.shared = new SharedHistory();
    // The ids of all scenes seen so far, in order.
    this.sceneStack = [];
    // History of consumable consumption.
    this.consumable = new ConsumableHistory();

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};
    
    /**
     * Adds the scene with the passed id to the scene stack.
     * Should be called during the load function of all scenes.
     * @param {string} id The id of the scene.
     */
    this.addToSceneStack = function(id) {
        if(that.sceneStack.length < 1) {
            that.sceneStack.push(id);
            return;
        }
        var lastScene = that.sceneStack[that.sceneStack.length-1];
        if(lastScene === id) {
            return;
        }
        that.sceneStack.push(id);
    };
    
    /**
     * Returns true if the shared storage has the specified key, false otherwise.
     * @param {string} key They key to look for.
     * @returns {boolean} True if the key exists, false otherwise.
     */
    this.hasShared = function(key) {
        return that.shared.hasOwnProperty(key);
    };
    
    /**
     * Gets the shared storage.
     * @returns {object} Shared scene storage.
     */
    this.getShared = function() {
        return that.shared;
    };
    
    /**
     * Sets the shared storage object.
     * @param {object} newStorage The new storage object.
     */
    this.setShared = function(newStorage) {
        that.shared = newStorage;
    };

    /**
     * Returns true if the passed scene has a history object, false otherwise.
     * @param {string} sceneName The name of the scene to check the history for.
     */
    this.exists = function (sceneName) {
        return that.scenes.hasOwnProperty(sceneName);
    };
    
    /**
     * Returns whether or not the passed scene's history object has the specified key.
     * @param {string} sceneName The scene of which to check the history.
     * @param {string} key The key to look for.
     */
    this.has = function(sceneName, key) {
        return that.exists(sceneName) && that.scenes[sceneName].hasOwnProperty(key);
    };

    /**
     * Fetches the history object of the passed scene and returns it.
     * Returns an empty object if the scene does not have a history object associated with it.
     * @param {string} sceneName The name of the scene to get the history for.
     */
    this.get = function (sceneName) {
        if(that.exists(sceneName)){
            return that.scenes[sceneName];
        }
        return {};
    };
    
    /**
     * Gets a specified key of a scene.
     * @param {string} sceneName The scene name for which to get the key.
     * @param {string} key The key to get.
     * @returns {object} The resulting object or undefined if the key has not yet been set.
     */
    this.getKey = function(sceneName, key) {
        return that.get(sceneName)[key];
    };
    
    /**
     * Sets the history object for the passed scene.
     * @param {string} sceneName The name of the scene to set the history for.
     * @param {object} history The new history object.
     */
    this.set = function (sceneName, history) {
        that.scenes[sceneName] = history;
    };
    
    /**
     * Sets a specified key of a scene.
     * @param {string} sceneName The scene name for which to set the key.
     * @param {string} key The key to set.
     * @param {object} setTo What to set the key to.
     */
    this.setKey = function(sceneName, key, setTo) {
        var history = that.get(sceneName);
        history[key] = setTo;
        that.set(sceneName, history);
    };
    
    /**
     * Unsets a specified key of a scene.
     * @param {string} sceneName The scene name for which to set the key.
     * @param {string} key The key to set.
     */
    this.unsetKey = function(sceneName, key) {
        var history = that.get(sceneName);
        delete history[key];
        that.set(sceneName, history);
    };
    
    /**
     * Performs the function if the specified key is not yet set,
     * then sets the key to setTo. Does nothing if the key is already set.
     * @param {string} sceneName The name of the scene.
     * @param {string} key The key to check for.
     * @param {anything} setTo What to set the key to after performing the function.
     * @param {type} fun The function that will be performed should the key not be set yet.
     */
    this.doIfNotSet = function(sceneName, key, setTo, fun) {
        if (!that.has(sceneName, key)) {
            fun();
            var history = that.get(sceneName);
            history[key] = setTo;
            global.kobold.history.set(sceneName, history);
        }
    };
    
    /**
     * Sets the specified key to setTo if not yet set.
     * @param {string} sceneName The name of the scene.
     * @param {string} key The key to check for.
     * @param {anything} setTo What to set the key to if not yet set.
     */
    this.setIfNotSet = function(sceneName, key, setTo) {
        that.doIfNotSet(sceneName, key, setTo, function(){});
    };

    /**
     * Loads this object from a type-stripped version thereof.
     * @param {History-stripped} history A History object, stripped of private members and all functions.
     */
    this.load = function (history) {
        that.scenes = history.scenes;
        that.shared.load(history.shared);
        that.sceneStack = history.sceneStack;
        that.consumable.load(history.consumable);
    };
}
